(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = global || self, factory(global['module-loader'] = {}));
}(this, (function (exports) { 'use strict';

    /**
     * Utils
     */
    const hasOwnProperty = Object.prototype.hasOwnProperty;
    function hasOwn(obj, key) {
        return hasOwnProperty.call(obj, key);
    }
    function warn(condition, msg, vm) {
        if (!condition) {
            // eslint-disable-next-line no-console
            console.warn(`[@vue-async] ${msg}`, vm);
        }
    }
    function error(condition, msg, vm) {
        if (!condition) {
            // eslint-disable-next-line no-console
            console.error(`[@vue-async] ${msg}`, vm);
        }
    }
    function print(title, content) {
        // eslint-disable-next-line no-console
        console.log(`%c${title}%c${content}`, 'background: #00d1b2; padding: 5px; color: #fff; border-radius: 5px 0 0 5px', 'background: #555; padding: 5px; color: #fff; border-radius: 0 5px 5px 0');
    }
    const toString = (x) => Object.prototype.toString.call(x);
    function isPlainObject(x) {
        return toString(x) === '[object Object]';
    }

    let currentVue = null;
    function setCurrentVue(vue) {
        currentVue = vue;
    }

    function sanitizeComponent(Component) {
        // If Component already sanitized
        if (Component.options && Component._Ctor === Component) {
            return Component;
        }
        if (!Component.options) {
            Component = currentVue.extend(Component);
            Component._Ctor = Component;
        }
        else {
            Component._Ctor = Component;
            Component.extendOptions = Component.options;
        }
        // For debugging purpose
        if (!Component.options.name && Component.options.__file) {
            Component.options.name = Component.options.__file;
        }
        return Component;
    }
    function getComponentName(Component) {
        const component = sanitizeComponent(Component);
        return component.options.name;
    }

    const isProduction = process.env.NODE_ENF === 'production';
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    function dynamicComponent (Vue, store) {
        return {
            /**
             * add component to store
             * @param component vue component
             * @param position position
             */
            add(component, position) {
                if (!(typeof component === 'object' || typeof component === 'function')) {
                    return error(isProduction, 'dynamicComponent function "add" only accept object or function component as param.');
                }
                store.commit('dynamicComponent/add', {
                    component,
                    position,
                });
            },
            /**
             * remove component from store
             * @param name component name
             * @param position position
             */
            remove(name, position) {
                if (typeof name !== 'string') {
                    return error(isProduction, 'dynamicComponent function "remoce" only accept string name of component as param.');
                }
                store.commit('dynamicComponent/remove', {
                    name,
                    position,
                });
            },
        };
    }

    const isProduction$1 = process.env.NODE_ENF === 'production';
    var dynamicComponentState = {
        namespaced: true,
        state: {
            GLOBAL: {},
        },
        mutations: {
            // 注册插槽组件
            add(state, { component, position = 'GLOBAL' }) {
                const _component = component.component || component;
                // 必须使用外部 Vue 对象，本地 import 对象不一致
                const componentName = component.name || getComponentName(_component);
                if (!componentName) {
                    return error(isProduction$1, 'component must be named for dynamic component.');
                }
                if (!state[position]) {
                    state[position] = {};
                }
                if (state[position][componentName]) {
                    return error(isProduction$1, `async component "${componentName}" has existed in the ${position}, ignored it.`);
                }
                state[position][componentName] = component;
            },
            // 移除注册插槽组件
            remove(state, { name, position = 'GLOBAL' }) {
                if (!state[position]) {
                    return error(isProduction$1, `position "${position}" does not  exist.`);
                }
                if (!state[position][name]) {
                    return error(isProduction$1, `async component "${name}" does not  exist in the ${position}.`);
                }
                delete state[position][name];
            },
        },
    };

    const isProduction$2 = process.env.NODE_ENF === 'production';
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    function eventBus (Vue) {
        let events = {};
        return {
            /**
             * 发射事件
             * @param {string} eventName 事件名称
             * @param {any} payload 事件载荷
             */
            emit(eventName, payload) {
                if (!events[eventName]) {
                    return warn(isProduction$2, `event "${eventName}" is triggered, but no listeners，nothing will be happening.`);
                }
                for (const func of events[eventName].values()) {
                    func(payload);
                }
            },
            /**
             * 事件监听
             * @param eventName 事件名称
             * @param handler 处理事件的方法
             */
            on(eventName, handler) {
                if (typeof handler === 'function') {
                    if (!events[eventName])
                        events[eventName] = new Set();
                    events[eventName].add(handler);
                }
                else {
                    error(isProduction$2, `handler must be a function, but recived ${typeof handler}`);
                }
            },
            /**
             * 取消事件监听
             * @param eventName 事件名称
             * @param handler 处理事件的方法
             */
            off(eventName, handler) {
                if (!handler) {
                    return warn(!isProduction$2, 'handler must be a valid event function');
                }
                else if (!events.hasOwnProperty(eventName)) {
                    return error(isProduction$2, `no event name of "${eventName}" is on listening`);
                }
                events[eventName].delete(handler);
            },
            /**
             * 清理事件总线
             */
            clear() {
                events = {};
            },
            /**
             * 获取当前事件总线详情
             */
            getEvents() {
                return events;
            },
        };
    }

    const isProduction$3 = "development" === 'production';
    var moduleLoader = (Vue, status) => {
        return function loader(modules, { __isArray__ = false } = {}) {
            status.current = false;
            if (Array.isArray(modules) && modules.length) {
                return Promise.all(modules.map(module => loader.call(this, module, { __isArray__: true }))).then(() => {
                    status.current = true;
                });
            }
            else if (typeof modules === 'object' && Object.getPrototypeOf(modules) === Object.prototype) {
                /** 通过模块清单加载模块 */
                const promiseAll = [];
                for (const moduleName in modules) {
                    if (!window[moduleName]) {
                        promiseAll.push(new Promise(resolve => {
                            const script = document.createElement('script');
                            script.src = modules[moduleName];
                            script.onload = () => {
                                if (window[moduleName]) {
                                    typeof window[moduleName] === 'function'
                                        ? window[moduleName].call(this, Vue)
                                        : window[moduleName].default.call(this, Vue);
                                    {
                                        print(moduleName, 'loaded');
                                    }
                                }
                                else {
                                    warn(isProduction$3, `module "${moduleName}" loaded failed，ignored it.`);
                                }
                                resolve();
                            };
                            script.onerror = () => {
                                warn(isProduction$3, `module "${moduleName}" had a problem to create, ignored it.`);
                                resolve();
                            };
                            document.body.appendChild(script);
                        }));
                    }
                }
                return Promise.all(promiseAll).then(() => {
                    if (!__isArray__) {
                        status.current = true;
                    }
                });
            }
            else if (typeof modules === 'function') {
                /** 通过模块函数加载模块 */
                const result = modules.call(this, Vue);
                if (result && result instanceof Promise) {
                    return result.then(() => {
                        if (!__isArray__) {
                            status.current = true;
                        }
                    });
                }
                else {
                    if (!__isArray__) {
                        status.current = true;
                    }
                    return Promise.resolve();
                }
            }
            else {
                error(isProduction$3, '模块加载方法只接受模块列表对象或者模块函数对象作为参数。');
                if (!__isArray__) {
                    status.current = true;
                }
                return Promise.resolve();
            }
        };
    };

    function install(Vue, options = {}) {
        if (install.installed)
            return;
        install.installed = true;
        // 用于监听模块加载的情况，处理路由404问题
        const vm = new Vue({
            data() {
                return {
                    status: {
                        current: false,
                    },
                };
            },
        });
        Object.defineProperties(Vue.prototype, {
            $eventBus: {
                value: eventBus(),
                writable: false,
            },
            $moduleLoader: {
                value: moduleLoader(Vue, vm.status),
                writable: false,
            },
        });
        const { store, router } = options;
        // store
        if (store) {
            store.registerModule('dynamicComponent', dynamicComponentState);
            // define $dynamicComponent
            Object.defineProperty(Vue.prototype, '$dynamicComponent', {
                value: dynamicComponent(Vue, store),
                writable: false,
            });
        }
        // router
        // 解决动态路由404问题
        if (router) {
            const resolveRoute = (to, next) => {
                const fullPath = to.redirectedFrom || to.fullPath;
                const { resolved, location } = router.resolve(fullPath);
                // 在加载完组件后resolve的地址与原来要跳转的地址不一致时跳转
                // 以免造成404死循环
                if (resolved.name !== to.name) {
                    next(location);
                }
                else {
                    next();
                }
            };
            router.beforeEach((to, from, next) => {
                if (!to.name || to.name === '404' || to.name === 'page-not-found') {
                    // 模块已经被加载完成, 但由于在其之前添加 beforeEach 阻止时间过长，vm.$watch还没开始监听
                    if (vm.status.current) {
                        resolveRoute(to, next);
                    }
                    vm.$watch(() => vm.status.current, (newVal, oldVal) => {
                        // false => true
                        if (newVal && !oldVal) {
                            resolveRoute(to, next);
                        }
                    });
                }
                else {
                    next();
                }
            });
        }
        // Used to avoid multiple mixins being setup
        // when in dev mode and hot module reload
        // https://github.com/vuejs/vue/issues/5089#issuecomment-284260111
        if (Vue.$__module_loader_installed__)
            return;
        // eslint-disable-next-line @typescript-eslint/camelcase
        Vue.$__module_loader_installed__ = true;
        // 设置当前运行时 Vue 对象
        setCurrentVue(Vue);
        // 设置子模块中的运行时 Vue 对象与主框架一致
        // 或使用 '--inline-vue' 使用独立 Vue 对象
        if (!window.Vue) {
            window.Vue = Vue;
        }
        Vue.mixin({
            beforeCreate() {
                const options = this.$options;
                if (options.moduleLoader) {
                    options.moduleLoader.init(this, options.ssrContext);
                    this.$moduleLoadManager = Vue.observable(options.moduleLoader.framework);
                    // 初始化实例中的 modules
                    // 放在最后执行，module 入口可以使用 this.$moduleLoadManager中的功能
                    options.moduleLoader.initModules(this).finally(() => {
                        // 在没有模块被加载时，这个值必须要手动改成 true
                        // 使 router.beforeEach 正确执行
                        if (!vm.status.current) {
                            vm.status.current = true;
                        }
                    });
                }
                else {
                    this.$moduleLoadManager = (options.parent && options.parent.$moduleLoadManager) || this;
                }
            },
        });
    }

    class ModuleLoader {
        constructor({ modules, ...rest } = {}) {
            this.framework = {
                loaded: false,
                layouts: {},
            };
            this.__initModules__ = modules;
            this.__ExtraParams__ = rest;
        }
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        init(root, ssrContext) {
            this.__ExtraParams__ &&
                Object.entries(this.__ExtraParams__).map(([key, value]) => {
                    if (!hasOwn(this.framework, key)) {
                        this.framework[key] = value;
                    }
                    else {
                        warn("development" === 'production', `参数 ${key} 已在在`);
                    }
                });
            // 默认方法
            !hasOwn(this.framework, 'addRoutes') && (this.framework.addRoutes = this.initAddRoutes(root));
            !hasOwn(this.framework, 'addLayouts') && (this.framework.addLayouts = this.initAddLayouts(root));
        }
        initModules(root) {
            if (this.__initModules__) {
                return root.$moduleLoader(this.__initModules__).then(() => {
                    this.framework.loaded = true;
                });
            }
            else {
                this.framework.loaded = true;
                return Promise.resolve();
            }
        }
        initAddRoutes(root) {
            return (routes) => {
                if (root.$router) {
                    root.$router.addRoutes(routes);
                }
                else {
                    error("development" === 'production', 'vm.$router未定义');
                }
            };
        }
        initAddLayouts(root) {
            return (key, layout) => {
                if (typeof key === 'string') {
                    root.$set(this.framework.layouts, key, layout);
                }
                else if (isPlainObject(key)) {
                    // plain object
                    const layouts = key;
                    this.framework.layouts = Object.assign(this.framework.layouts, layouts);
                }
                else {
                    error("development" === 'production', 'addLayouts中支持字符串(key)或 Object(key/value)格式');
                }
            };
        }
    }
    ModuleLoader.install = install;
    // static installed = false
    ModuleLoader.version = '0.1.3-alpha.0';

    const install$1 = ModuleLoader.install;
    ModuleLoader.install = (Vue, options = {}) => {
        install$1.call(ModuleLoader, Vue, options);
    };
    if (typeof window !== 'undefined' && window.Vue) {
        window.Vue.use(ModuleLoader);
    }

    exports.default = ModuleLoader;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
